@@ -1,27 +1,31 @@
 #This file is part of Tryton.  The COPYRIGHT file at the top level of
 #this repository contains the full copyright notices and license terms.
 from decimal import Decimal
-import datetime
 from itertools import groupby, chain
 from functools import partial
+from sql import Table
+from sql.functions import Overlay, Position
+from sql.operators import Concat
+
 from trytond.model import Workflow, ModelView, ModelSQL, fields
 from trytond.modules.company import CompanyReport
 from trytond.wizard import Wizard, StateAction, StateView, StateTransition, \
     Button
-from trytond.backend import TableHandler
+from trytond import backend
 from trytond.pyson import If, Eval, Bool, PYSONEncoder, Id
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
 
 __all__ = ['Sale', 'SaleInvoice', 'SaleIgnoredInvoice', 'SaleRecreatedInvoice',
     'SaleLine', 'SaleLineTax', 'SaleLineIgnoredMove',
-    'SaleLineRecreatedMove', 'SaleReport', 'Template', 'Product',
-    'ShipmentOut', 'ShipmentOutReturn', 'Move', 'OpenCustomer',
+    'SaleLineRecreatedMove', 'SaleReport', 'OpenCustomer',
     'HandleShipmentExceptionAsk', 'HandleShipmentException',
     'HandleInvoiceExceptionAsk', 'HandleInvoiceException',
-    'ReturnSale']
+    'ReturnSaleStart', 'ReturnSale']
 __metaclass__ = PoolMeta
 
+_ZERO = Decimal(0)
+
 
 class Sale(Workflow, ModelSQL, ModelView):
     'Sale'
@@ -33,7 +37,7 @@ class Sale(Workflow, ModelSQL, ModelView):
             },
         domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
+                Eval('context', {}).get('company', -1)),
             ],
         depends=['state'], select=True)
     reference = fields.Char('Reference', readonly=True, select=True)
@@ -97,21 +101,21 @@ class Sale(Workflow, ModelSQL, ModelView):
     comment = fields.Text('Comment')
     untaxed_amount = fields.Function(fields.Numeric('Untaxed',
             digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_untaxed_amount')
+            depends=['currency_digits']), 'get_amount')
     untaxed_amount_cache = fields.Numeric('Untaxed Cache',
         digits=(16, Eval('currency_digits', 2)),
         readonly=True,
         depends=['currency_digits'])
     tax_amount = fields.Function(fields.Numeric('Tax',
             digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_tax_amount')
+            depends=['currency_digits']), 'get_amount')
     tax_amount_cache = fields.Numeric('Tax Cache',
         digits=(16, Eval('currency_digits', 2)),
         readonly=True,
         depends=['currency_digits'])
     total_amount = fields.Function(fields.Numeric('Total',
             digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']), 'get_total_amount')
+            depends=['currency_digits']), 'get_amount')
     total_amount_cache = fields.Numeric('Total Tax',
         digits=(16, Eval('currency_digits', 2)),
         readonly=True,
@@ -154,10 +158,10 @@ class Sale(Workflow, ModelSQL, ModelView):
             ('exception', 'Exception'),
             ], 'Shipment State', readonly=True, required=True)
     shipments = fields.Function(fields.One2Many('stock.shipment.out', None,
-        'Shipments'), 'get_shipments')
+        'Shipments'), 'get_shipments', searcher='search_shipments')
     shipment_returns = fields.Function(
         fields.One2Many('stock.shipment.out.return', None, 'Shipment Returns'),
-        'get_shipment_returns')
+        'get_shipment_returns', searcher='search_shipment_returns')
     moves = fields.Function(fields.One2Many('stock.move', None, 'Moves'),
         'get_moves')
 
@@ -169,7 +173,8 @@ class Sale(Workflow, ModelSQL, ModelView):
         cls._error_messages.update({
                 'invalid_method': ('Invalid combination of shipment and '
                     'invoicing methods on sale "%s".'),
-                'addresses_required': ('Invoice and Shipment addresses must be '
+                'addresses_required': (
+                    'Invoice and Shipment addresses must be '
                     'defined for the quotation of sale "%s".'),
                 'warehouse_required': ('Warehouse must be defined for the '
                     'quotation of sale "%s".'),
@@ -227,20 +232,36 @@ class Sale(Workflow, ModelSQL, ModelView):
     def __register__(cls, module_name):
         pool = Pool()
         SaleLine = pool.get('sale.line')
-        sale_line_invoice_line_table = 'sale_line_invoice_lines_rel'
+        TableHandler = backend.get('TableHandler')
+        sale_line_invoice_line_table_name = 'sale_line_invoice_lines_rel'
         Move = pool.get('stock.move')
         cursor = Transaction().cursor
+        model_data = Table('ir_model_data')
+        model_field = Table('ir_model_field')
+        sql_table = cls.__table__()
+
         # Migration from 1.2: packing renamed into shipment
-        cursor.execute("UPDATE ir_model_data "
-                "SET fs_id = REPLACE(fs_id, 'packing', 'shipment') "
-                "WHERE fs_id like '%%packing%%' AND module = %s",
-                (module_name,))
-        cursor.execute("UPDATE ir_model_field "
-                "SET relation = REPLACE(relation, 'packing', 'shipment'), "
-                    "name = REPLACE(name, 'packing', 'shipment') "
-                "WHERE (relation like '%%packing%%' "
-                    "OR name like '%%packing%%') AND module = %s",
-                (module_name,))
+        cursor.execute(*model_data.update(
+                columns=[model_data.fs_id],
+                values=[Overlay(model_data.fs_id, 'shipment',
+                        Position('packing', model_data.fs_id),
+                        len('packing'))],
+                where=model_data.fs_id.like('%packing%')
+                & (model_data.module == module_name)))
+        cursor.execute(*model_field.update(
+                columns=[model_field.relation],
+                values=[Overlay(model_field.relation, 'shipment',
+                        Position('packing', model_field.relation),
+                        len('packing'))],
+                where=model_field.relation.like('%packing%')
+                & (model_field.module == module_name)))
+        cursor.execute(*model_field.update(
+                columns=[model_field.name],
+                values=[Overlay(model_field.name, 'shipment',
+                        Position('packing', model_field.name),
+                        len('packing'))],
+                where=model_field.name.like('%packing%')
+                & (model_field.module == module_name)))
         table = TableHandler(cursor, cls, module_name)
         table.column_rename('packing_state', 'shipment_state')
         table.column_rename('packing_method', 'shipment_method')
@@ -249,9 +270,10 @@ class Sale(Workflow, ModelSQL, ModelView):
         super(Sale, cls).__register__(module_name)
 
         # Migration from 1.2
-        cursor.execute("UPDATE " + cls._table + " "
-                "SET invoice_method = 'shipment' "
-                "WHERE invoice_method = 'packing'")
+        cursor.execute(*sql_table.update(
+                columns=[sql_table.invoice_method],
+                values=['shipment'],
+                where=sql_table.invoice_method == 'packing'))
 
         table = TableHandler(cursor, cls, module_name)
         # Migration from 2.2
@@ -260,27 +282,30 @@ class Sale(Workflow, ModelSQL, ModelView):
         # state confirmed splitted into confirmed and processing
         if (TableHandler.table_exist(cursor, SaleLine._table)
                 and TableHandler.table_exist(cursor,
-                    sale_line_invoice_line_table)
+                    sale_line_invoice_line_table_name)
                 and TableHandler.table_exist(cursor, Move._table)):
+            sale_line = SaleLine.__table__()
+            sale_line_invoice_line = \
+                Table(sale_line_invoice_line_table_name)
+            move = Move.__table__()
             # Wrap subquery inside an other inner subquery because MySQL syntax
             # doesn't allow update a table and select from the same table in a
             # subquery.
-            cursor.execute('UPDATE "%s" '
-                "SET state = 'processing' "
-                'WHERE id IN ('
-                    'SELECT id '
-                    'FROM ('
-                        'SELECT s.id '
-                        'FROM "%s" AS s '
-                        'INNER JOIN "%s" AS l ON l.sale = s.id '
-                        'LEFT JOIN "%s" AS li ON li.sale_line = l.id '
-                        'LEFT JOIN "%s" AS m ON m.origin = \'%s,\' || l.id '
-                        "WHERE s.state = 'confirmed' "
-                            'AND (li.id IS NOT NULL '
-                                'OR m.id IS NOT NULL)) AS foo)'
-                % (cls._table, cls._table, SaleLine._table,
-                    sale_line_invoice_line_table, Move._table,
-                    SaleLine.__name__))
+            sub_query = sql_table.join(sale_line,
+                condition=sale_line.sale == sql_table.id
+                ).join(sale_line_invoice_line, 'LEFT',
+                    condition=sale_line_invoice_line.sale_line == sale_line.id
+                    ).join(move, 'LEFT',
+                        condition=(move.origin == Concat(SaleLine.__name__,
+                                sale_line.id))
+                        ).select(sql_table.id,
+                            where=(sql_table.state == 'confirmed')
+                            & ((sale_line_invoice_line.id != None)
+                                | (move.id != None)))
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.state],
+                    values=['processing'],
+                    where=sql_table.id.in_(sub_query.select(sub_query.id))))
 
         # Add index on create_date
         table = TableHandler(cursor, cls, module_name)
@@ -427,22 +452,11 @@ class Sale(Workflow, ModelSQL, ModelView):
             res['total_amount'] = self.currency.round(res['total_amount'])
         return res
 
-    def get_untaxed_amount(self, name):
-        if (self.state in self._states_cached
-                and self.untaxed_amount_cache is not None):
-            return self.untaxed_amount_cache
-        amount = sum((l.amount for l in self.lines if l.type == 'line'),
-            Decimal(0))
-        return self.currency.round(amount)
-
-    def get_tax_amount(self, name):
+    def get_tax_amount(self):
         pool = Pool()
         Tax = pool.get('account.tax')
         Invoice = pool.get('account.invoice')
 
-        if (self.state in self._states_cached
-                and self.tax_amount_cache is not None):
-            return self.tax_amount_cache
         context = self.get_tax_context()
         taxes = {}
         for line in self.lines:
@@ -458,15 +472,39 @@ class Sale(Workflow, ModelSQL, ModelView):
                     taxes[key] = val['amount']
                 else:
                     taxes[key] += val['amount']
-        amount = sum((self.currency.round(taxes[key]) for key in taxes),
-            Decimal(0))
-        return self.currency.round(amount)
+        return sum((self.currency.round(tax) for tax in taxes.values()), _ZERO)
 
-    def get_total_amount(self, name):
-        if (self.state in self._states_cached
-                and self.total_amount_cache is not None):
-            return self.total_amount_cache
-        return self.currency.round(self.untaxed_amount + self.tax_amount)
+    @classmethod
+    def get_amount(cls, sales, names):
+        untaxed_amount = {}
+        tax_amount = {}
+        total_amount = {}
+
+        for sale in sales:
+            if (sale.state in cls._states_cached
+                    and sale.untaxed_amount_cache is not None
+                    and sale.tax_amount_cache is not None
+                    and sale.total_amount_cache is not None):
+                untaxed_amount[sale.id] = sale.untaxed_amount_cache
+                tax_amount[sale.id] = sale.tax_amount_cache
+                total_amount[sale.id] = sale.total_amount_cache
+            else:
+                untaxed_amount[sale.id] = sum(
+                    (line.amount for line in sale.lines
+                        if line.type == 'line'), _ZERO)
+                tax_amount[sale.id] = sale.get_tax_amount()
+                total_amount[sale.id] = (
+                    untaxed_amount[sale.id] + tax_amount[sale.id])
+
+        result = {
+            'untaxed_amount': untaxed_amount,
+            'tax_amount': tax_amount,
+            'total_amount': total_amount,
+            }
+        for key in result.keys():
+            if key not in names:
+                del result[key]
+        return result
 
     def get_invoice_state(self):
         '''
@@ -509,6 +547,19 @@ class Sale(Workflow, ModelSQL, ModelView):
     get_shipments = get_shipments_returns('stock.shipment.out')
     get_shipment_returns = get_shipments_returns('stock.shipment.out.return')
 
+    def search_shipments_returns(model_name):
+        '''
+        Search on shipments or returns
+        '''
+        def method(self, name, clause):
+            return [('lines.moves.shipment.id',) + tuple(clause[1:])
+                + (model_name,)]
+        return classmethod(method)
+
+    search_shipments = search_shipments_returns('stock.shipment.out')
+    search_shipment_returns = search_shipments_returns(
+        'stock.shipment.out.return')
+
     def get_moves(self, name):
         return [m.id for l in self.lines for m in l.moves]
 
@@ -659,7 +710,6 @@ class Sale(Workflow, ModelSQL, ModelView):
             return Invoice(
                 company=self.company,
                 type=invoice_type,
-                reference=self.reference,
                 journal=journal,
                 party=self.party,
                 invoice_address=self.invoice_address,
@@ -686,7 +736,9 @@ class Sale(Workflow, ModelSQL, ModelView):
             return
 
         invoice = self._get_invoice_sale(invoice_type)
-        invoice.lines = list(chain.from_iterable(invoice_lines.itervalues()))
+        invoice.lines = ((list(invoice.lines)
+                if hasattr(invoice, 'lines') else [])
+            + list(chain.from_iterable(invoice_lines.itervalues())))
         invoice.save()
 
         with Transaction().set_user(0, set_context=True):
@@ -752,7 +804,6 @@ class Sale(Workflow, ModelSQL, ModelView):
             values = {
                 'customer': self.party.id,
                 'delivery_address': self.shipment_address.id,
-                'reference': self.reference,
                 'company': self.company.id,
                 }
             values.update(dict(key))
@@ -876,9 +927,7 @@ class SaleLine(ModelSQL, ModelView):
     _rec_name = 'description'
     sale = fields.Many2One('sale.sale', 'Sale', ondelete='CASCADE',
         select=True)
-    sequence = fields.Integer('Sequence',
-        order_field='(%(table)s.sequence IS NULL) %(order)s, '
-        '%(table)s.sequence %(order)s')
+    sequence = fields.Integer('Sequence')
     type = fields.Selection([
         ('line', 'Line'),
         ('subtotal', 'Subtotal'),
@@ -924,7 +973,6 @@ class SaleLine(ModelSQL, ModelView):
             'locations': If(Bool(Eval('_parent_sale', {}).get('warehouse')),
                 [Eval('_parent_sale', {}).get('warehouse', 0)], []),
             'stock_date_end': Eval('_parent_sale', {}).get('sale_date'),
-            'salable': True,
             'stock_skip_warehouse': True,
             }, depends=['type'])
     product_uom_category = fields.Function(
@@ -985,7 +1033,8 @@ class SaleLine(ModelSQL, ModelView):
         super(SaleLine, cls).__setup__()
         cls._order.insert(0, ('sequence', 'ASC'))
         cls._error_messages.update({
-                'customer_location_required': ('Sale "%(sale)s" is missing the '
+                'customer_location_required': (
+                    'Sale "%(sale)s" is missing the '
                     'customer location in line "%(line)s".'),
                 'missing_account_revenue': ('Product "%(product)s" of sale '
                     '%(sale)s misses a revenue account.'),
@@ -995,18 +1044,27 @@ class SaleLine(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        super(SaleLine, cls).__register__(module_name)
+        TableHandler = backend.get('TableHandler')
         cursor = Transaction().cursor
+        sql_table = cls.__table__()
+        super(SaleLine, cls).__register__(module_name)
         table = TableHandler(cursor, cls, module_name)
 
         # Migration from 1.0 comment change into note
         if table.column_exist('comment'):
-            cursor.execute('UPDATE "' + cls._table + '" SET note = comment')
+            cursor.execute(*sql_table.update(
+                    columns=[sql_table.note],
+                    values=[sql_table.comment]))
             table.drop_column('comment', exception=True)
 
         # Migration from 2.4: drop required on sequence
         table.not_null_action('sequence', action='remove')
 
+    @staticmethod
+    def order_sequence(tables):
+        table, _ = tables[None]
+        return [table.sequence == None, table.sequence]
+
     @staticmethod
     def default_type():
         return 'line'
@@ -1251,6 +1309,7 @@ class SaleLine(ModelSQL, ModelView):
         invoice_line.product = self.product
         invoice_line.unit_price = self.unit_price
         invoice_line.taxes = self.taxes
+        invoice_line.invoice_type = invoice_type
         if self.product:
             invoice_line.account = self.product.account_revenue_used
             if not invoice_line.account:
@@ -1302,7 +1361,7 @@ class SaleLine(ModelSQL, ModelView):
         else:
             quantity = 0.0
             for invoice_line in self.invoice_lines:
-                if invoice_line.invoice.state in ('posted', 'paid'):
+                if invoice_line.invoice.state == 'paid':
                     quantity += Uom.compute_qty(invoice_line.unit,
                         invoice_line.quantity, self.unit)
 
@@ -1368,300 +1427,6 @@ class SaleReport(CompanyReport):
     __name__ = 'sale.sale'
 
 
-class Template:
-    __name__ = 'product.template'
-    salable = fields.Boolean('Salable', states={
-            'readonly': ~Eval('active', True),
-            }, depends=['active'])
-    sale_uom = fields.Many2One('product.uom', 'Sale UOM', states={
-            'readonly': ~Eval('active', True),
-            'invisible': ~Eval('salable', False),
-            'required': Eval('salable', False),
-            },
-        domain=[
-            ('category', '=', Eval('default_uom_category')),
-            ],
-        on_change_with=['default_uom', 'sale_uom', 'salable'],
-        depends=['active', 'salable', 'default_uom_category'])
-    delivery_time = fields.Integer('Delivery Time', states={
-            'readonly': ~Eval('active', True),
-            'invisible': ~Eval('salable', False),
-            'required': Eval('salable', False),
-            },
-        depends=['active', 'salable'],
-        help='In number of days')
-
-    @classmethod
-    def __setup__(cls):
-        super(Template, cls).__setup__()
-        required = ~Eval('account_category', False) & Eval('salable', False)
-        if not cls.account_revenue.states.get('required'):
-            cls.account_revenue.states['required'] = required
-        else:
-            cls.account_revenue.states['required'] = (
-                    cls.account_revenue.states['required'] | required)
-        if 'account_category' not in cls.account_revenue.depends:
-            cls.account_revenue.depends.append('account_category')
-        if 'salable' not in cls.account_revenue.depends:
-            cls.account_revenue.depends.append('salable')
-
-    @staticmethod
-    def default_salable():
-        return True if Transaction().context.get('salable') else False
-
-    @staticmethod
-    def default_delivery_time():
-        return 0
-
-    def on_change_with_sale_uom(self):
-        if self.default_uom:
-            if self.sale_uom:
-                if self.default_uom.category == self.sale_uom.category:
-                    return self.sale_uom.id
-                else:
-                    return self.default_uom.id
-            else:
-                return self.default_uom.id
-
-
-class Product:
-    __name__ = 'product.product'
-
-    @staticmethod
-    def get_sale_price(products, quantity=0):
-        '''
-        Return the sale price for products and quantity.
-        It uses if exists from the context:
-            uom: the unit of measure
-            currency: the currency id for the returned price
-        '''
-        pool = Pool()
-        Uom = pool.get('product.uom')
-        User = pool.get('res.user')
-        Currency = pool.get('currency.currency')
-        Date = pool.get('ir.date')
-
-        today = Date.today()
-        prices = {}
-
-        uom = None
-        if Transaction().context.get('uom'):
-            uom = Uom(Transaction().context.get('uom'))
-
-        currency = None
-        if Transaction().context.get('currency'):
-            currency = Currency(Transaction().context.get('currency'))
-
-        user = User(Transaction().user)
-
-        for product in products:
-            prices[product.id] = product.list_price
-            if uom:
-                prices[product.id] = Uom.compute_price(
-                    product.default_uom, prices[product.id], uom)
-            if currency and user.company:
-                if user.company.currency != currency:
-                    date = Transaction().context.get('sale_date') or today
-                    with Transaction().set_context(date=date):
-                        prices[product.id] = Currency.compute(
-                            user.company.currency, prices[product.id],
-                            currency, round=False)
-        return prices
-
-    def compute_delivery_date(self, date=None):
-        '''
-        Compute the delivery date a the given date
-        '''
-        Date = Pool().get('ir.date')
-
-        if not date:
-            date = Date.today()
-        return date + datetime.timedelta(self.delivery_time)
-
-
-class ShipmentOut:
-    __name__ = 'stock.shipment.out'
-
-    @classmethod
-    def __setup__(cls):
-        super(ShipmentOut, cls).__setup__()
-        cls._error_messages.update({
-                'reset_move': 'You cannot reset to draft a move generated '
-                    'by a sale.',
-            })
-
-    @classmethod
-    def write(cls, shipments, vals):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        super(ShipmentOut, cls).write(shipments, vals)
-
-        if 'state' in vals and vals['state'] in ('done', 'cancel'):
-            sales = []
-            move_ids = []
-            for shipment in shipments:
-                move_ids.extend([x.id for x in shipment.outgoing_moves])
-
-            with Transaction().set_user(0, set_context=True):
-                sale_lines = SaleLine.search([
-                        ('moves', 'in', move_ids),
-                        ])
-                if sale_lines:
-                    sales = list(set(l.sale for l in sale_lines))
-                    Sale.process(sales)
-
-    @classmethod
-    @ModelView.button
-    @Workflow.transition('draft')
-    def draft(cls, shipments):
-        SaleLine = Pool().get('sale.line')
-        for shipment in shipments:
-            for move in shipment.outgoing_moves:
-                if (move.state == 'cancel'
-                        and isinstance(move.origin, SaleLine)):
-                    cls.raise_user_error('reset_move')
-
-        return super(ShipmentOut, cls).draft(shipments)
-
-
-class ShipmentOutReturn:
-    __name__ = 'stock.shipment.out.return'
-
-    @classmethod
-    def __setup__(cls):
-        super(ShipmentOutReturn, cls).__setup__()
-        cls._error_messages.update({
-                'reset_move': 'You cannot reset to draft a move generated '
-                    'by a sale.',
-            })
-
-    @classmethod
-    def write(cls, shipments, vals):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        super(ShipmentOutReturn, cls).write(shipments, vals)
-
-        if 'state' in vals and vals['state'] == 'received':
-            sales = []
-            move_ids = []
-            for shipment in shipments:
-                move_ids.extend([x.id for x in shipment.incoming_moves])
-
-            with Transaction().set_user(0, set_context=True):
-                sale_lines = SaleLine.search([
-                        ('moves', 'in', move_ids),
-                        ])
-                if sale_lines:
-                    for sale_line in sale_lines:
-                        if sale_line.sale not in sales:
-                            sales.append(sale_line.sale)
-
-                    sales = Sale.browse([s.id for s in sales])
-                    Sale.process(sales)
-
-    @classmethod
-    @ModelView.button
-    @Workflow.transition('draft')
-    def draft(cls, shipments):
-        SaleLine = Pool().get('sale.line')
-        for shipment in shipments:
-            for move in shipment.incoming_moves:
-                if (move.state == 'cancel'
-                        and isinstance(move.origin, SaleLine)):
-                    cls.raise_user_error('reset_move')
-
-        return super(ShipmentOutReturn, cls).draft(shipments)
-
-
-class Move:
-    __name__ = 'stock.move'
-    sale = fields.Function(fields.Many2One('sale.sale', 'Sale', select=True),
-        'get_sale', searcher='search_sale')
-    sale_exception_state = fields.Function(fields.Selection([
-        ('', ''),
-        ('ignored', 'Ignored'),
-        ('recreated', 'Recreated'),
-        ], 'Exception State'), 'get_sale_exception_state')
-
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().cursor
-
-        super(Move, cls).__register__(module_name)
-
-        table = TableHandler(cursor, cls, module_name)
-
-        # Migration from 2.6: remove sale_line
-        if table.column_exist('sale_line'):
-            cursor.execute('UPDATE "' + cls._table + '" '
-                'SET origin = \'sale.line,\' || sale_line '
-                'WHERE sale_line IS NOT NULL')
-            table.drop_column('sale_line')
-
-    @classmethod
-    def _get_origin(cls):
-        models = super(Move, cls)._get_origin()
-        models.append('sale.line')
-        return models
-
-    def get_sale(self, name):
-        SaleLine = Pool().get('sale.line')
-        if isinstance(self.origin, SaleLine):
-            return self.origin.sale.id
-
-    @classmethod
-    def search_sale(cls, name, clause):
-        return [('origin.' + name,) + tuple(clause[1:]) + ('sale.line',)]
-
-    def get_sale_exception_state(self, name):
-        SaleLine = Pool().get('sale.line')
-        if not isinstance(self.origin, SaleLine):
-            return ''
-        if self in self.origin.moves_recreated:
-            return 'recreated'
-        if self in self.origin.moves_ignored:
-            return 'ignored'
-        return ''
-
-    @classmethod
-    def write(cls, moves, vals):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        super(Move, cls).write(moves, vals)
-        if 'state' in vals and vals['state'] in ('cancel',):
-            with Transaction().set_user(0, set_context=True):
-                sale_lines = SaleLine.search([
-                        ('moves', 'in', [m.id for m in moves]),
-                        ])
-                if sale_lines:
-                    sales = list(set(l.sale for l in sale_lines))
-                    Sale.process(sales)
-
-    @classmethod
-    def delete(cls, moves):
-        pool = Pool()
-        Sale = pool.get('sale.sale')
-        SaleLine = pool.get('sale.line')
-
-        with Transaction().set_user(0, set_context=True):
-            sale_lines = SaleLine.search([
-                    ('moves', 'in', [m.id for m in moves]),
-                    ])
-
-        super(Move, cls).delete(moves)
-
-        if sale_lines:
-            sales = list(set(l.sale for l in sale_lines))
-            with Transaction().set_user(0, set_context=True):
-                Sale.process(sales)
-
-
 class OpenCustomer(Wizard):
     'Open Customers'
     __name__ = 'sale.open_customer'
@@ -1672,8 +1437,11 @@ class OpenCustomer(Wizard):
         pool = Pool()
         ModelData = pool.get('ir.model.data')
         Wizard = pool.get('ir.action.wizard')
+        Sale = pool.get('sale.sale')
         cursor = Transaction().cursor
-        cursor.execute("SELECT DISTINCT(party) FROM sale_sale")
+        sale = Sale.__table__()
+
+        cursor.execute(*sale.select(sale.party, group_by=sale.party))
         customer_ids = [line[0] for line in Transaction().cursor.fetchall()]
         action['pyson_domain'] = PYSONEncoder().encode(
             [('id', 'in', customer_ids)])
@@ -1703,11 +1471,15 @@ class HandleShipmentExceptionAsk(ModelView):
     @classmethod
     def __register__(cls, module_name):
         cursor = Transaction().cursor
+        model = Table('ir_model')
         # Migration from 1.2: packing renamed into shipment
-        cursor.execute("UPDATE ir_model "
-                "SET model = REPLACE(model, 'packing', 'shipment') "
-                "WHERE model like '%%packing%%' AND module = %s",
-                (module_name,))
+        cursor.execute(*model.update(
+                columns=[model.model],
+                values=[Overlay(model.model, 'shipment',
+                        Position('packing', model.model),
+                        len('packing'))],
+                where=model.model.like('%packing%')
+                & (model.module == module_name)))
         super(HandleShipmentExceptionAsk, cls).__register__(module_name)
 
 
@@ -1830,17 +1602,31 @@ class HandleInvoiceException(Wizard):
         return 'end'
 
 
-class ReturnSale(Wizard):
-    __name__ = 'sale.return_sale'
-    start_state = 'make_return'
-    make_return = StateTransition()
+class ReturnSaleStart(ModelView):
+    'Return Sale'
+    __name__ = 'sale.return_sale.start'
 
-    def transition_make_return(self):
+
+class ReturnSale(Wizard):
+    'Return Sale'
+    __name__ = 'sale.return_sale'
+    start = StateView('sale.return_sale.start',
+        'sale.return_sale_start_view_form', [
+            Button('Cancel', 'end', 'tryton-cancel'),
+            Button('Return', 'return_', 'tryton-ok', default=True),
+            ])
+    return_ = StateAction('sale.act_sale_form')
+
+    def do_return_(self, action):
         Sale = Pool().get('sale.sale')
 
-        sale = Sale(Transaction().context['active_id'])
-        new_sale, = Sale.copy([sale])
-        for new_line in new_sale.lines:
-            new_line.quantity *= -1
-            new_line.save()
-        return 'end'
+        sales = Sale.browse(Transaction().context['active_ids'])
+        return_sales = Sale.copy(sales)
+        for sale in return_sales:
+            for line in sale.lines:
+                line.quantity *= -1
+                line.save()
+        data = {'res_id': [s.id for s in return_sales]}
+        if len(return_sales) == 1:
+            action['views'].reverse()
+        return action, data
